
define(['../../../views/request', '../../../models/request', 'core_models/user'],
  function(RequestView, RequestModel, UserModel) {

  describe('Request view', function() {

    var view,
        model,
        user;

    beforeEach(function() {

      user = new UserModel({
        role: 'admin'
      });

      model = new RequestModel({
        name: 'test_csr'
      });

      view = new RequestView({
        model: model
      });

      view.core.user = user;
    });

    afterEach(function() {
      view.core.user = {};
    });

    it('should have the correct CSS when the request is approvable', function() {
      model.set('approvable', true);

      view.render();

      expect(view.$el.hasClass('certificate-valid-row')).toBeTruthy();
    });

    it('should have the correct CSS when the request is not approvable', function() {
      model.set('approvable', false);

      view.render();

      expect(view.$el.hasClass('certificate-invalid-row')).toBeTruthy();
    });

    describe('UI elements', function() {

      describe('for an actionable request', function() {

        beforeEach(function() {
          model.set('approvable', true);
          model.set('request_open', true);

          view.render();
        });

        it('should include an accept button', function() {
          expect(view.ui.accept_button.length).toBe(1);
        });

        it('should include a reject button', function() {
          expect(view.ui.reject_button.length).toBe(1);
        });
      });

      describe('for an unactionable request', function() {
        it('should show an appropriate message', function() {
          view.render();

          expect(view.el.innerHTML).toMatch(/Cannot accept or reject this request/);
        });
      });

    });

    describe('actioning a request', function() {

      var server;

      beforeEach(function() {
        server = sinon.fakeServer.create();

        model.set('approvable', true);
        model.set('request_open', true);

        view.render();
      });

      afterEach(function() {
        server.restore();
      });

      describe('by accepting it', function() {

        it('should show a loading indicator while processing the request', function() {
          view.accept_csr();

          expect(view.ui.loading_indicator.html()).toBeTruthy();
        });

        it('should hide the action buttons while processing the request', function() {
          view.accept_csr();

          expect(view.ui.action_buttons.css('display')).toBe('none');
        });

        it('rerenders the view on success', function() {

          server.respondWith([200, {}, 'OK']);

          view.accept_csr();

          server.respond();

          expect(view.el.innerHTML).toMatch(/Node request accepted/);
        });

        describe('on failure', function() {

          beforeEach(function() {
            server.respondWith([500, {}, 'ERROR']);

            view.accept_csr();

            server.respond();
          });

          it('hides the loading indicator', function() {
              expect(view.ui.loading_indicator.html()).toBeFalsy();
          });

          it('shows the action buttons', function() {
              expect(view.ui.action_buttons.css('display')).toBeTruthy();
          });

          it('displays an error message', function() {
            expect(view.ui.action_message.text()).toMatch(/Could not process this request/);
          });
        });
      });

      describe('by rejecting it', function() {

        it('should show a loading indicator while processing the request', function() {
          view.reject_csr();

          expect(view.ui.loading_indicator.html()).toBeTruthy();
        });

        it('should hide the action buttons while processing the request', function() {
          view.reject_csr();

          expect(view.ui.action_buttons.css('display')).toBe('none');
        });

        it('rerenders the view on success', function() {

          server.respondWith([200, {}, 'OK']);

          view.reject_csr();

          server.respond();

          expect(view.el.innerHTML).toMatch(/Node request rejected/);
        });

        describe('on failure', function() {

          beforeEach(function() {
            server.respondWith([500, {}, 'ERROR']);

            view.reject_csr();

            server.respond();
          });

          it('hides the loading indicator', function() {
            expect(view.ui.loading_indicator.html()).toBeFalsy();
          });

          it('shows the action buttons', function() {
            expect(view.ui.action_buttons.css('display')).toBeTruthy();
          });

          it('displays an error message', function() {
            expect(view.ui.action_message.text()).toMatch(/Could not process this request/);
          });
        });
      });
    });

    describe('using user access control', function() {

      beforeEach(function() {
        model.set('approvable', true);
        model.set('request_open', true);
      });

      it('should not show action buttons to read-only users', function() {
        user.set('role', 'read-only');

        view.render();

        expect(view.ui.accept_button.length).toBe(0);
        expect(view.ui.reject_button.length).toBe(0);
      });

      it('should show action buttons to read-write users', function() {
        user.set('role', 'read-write');

        view.render();

        expect(view.ui.accept_button.length).toBe(1);
        expect(view.ui.reject_button.length).toBe(1);
      });

      it('show should action buttons to admin users', function() {
        user.set('role', 'admin');

        view.render();

        expect(view.ui.accept_button.length).toBe(1);
        expect(view.ui.reject_button.length).toBe(1);
      });
    });

  });

});